"""
BFH: a library for mapping schemas to other schemas.

"""
from __future__ import absolute_import

from .common import nullish, dedunder
from .interfaces import SchemaInterface, MappingInterface

from . import exceptions
from . import fields
from . import transformations

from .version import __version__

__all__ = [
    "Schema",
    "Mapping",
    "exceptions",
    "fields",
    "transformations",
]


class Schema(SchemaInterface):
    """
    A base class for defining your schemas:

    Declare the shape of an object you expect to handle.

    Just inherit this and add some fields:

        class Animal(Schema):
            name = UnicodeField()
            type = UnicodeField()
            legs = IntegerField()
            noise = UnicodeField()

    """
    def __init__(self, *args, **kwargs):
        """
        Args:
           Pass a dictionary a single positional argument and it will be
           tranformed into kwargs.

        Kwargs:
           Values to assign to fields in the schema. Unknown names are ignored.
        """
        # when a dict is passed as positional argument, transform to kwargs
        if len(args) == 1 and isinstance(args[0], dict):
            return self.__init__(**dict(args[0], **kwargs))

        # stash raw kwargs for downstream
        self._raw_input = kwargs

        # init any subschemas
        for k, v in self._fields.items():
            if isinstance(v, fields.Subschema):
                setattr(self, k, v.subschema_class())

        # init values passed as kwargs
        for k, v in kwargs.items():
            if k.startswith("__"):
                k = k.strip("_")

            if k in self._field_names:
                setattr(self, k, v)

    def __setattr__(self, name, value):
        name = dedunder(name)
        object.__setattr__(self, name, value)

    def __getattr__(self, name):
        name = dedunder(name)
        return object.__getattribute__(self, name)

    def serialize(self, implicit_nulls=True):
        """
        Represent this schema as a dictionary.

        Handy for dumping to json or some other further serialization format,
        or splatting into an object constructor.

        Kwargs:
            implicit_nulls (Bool) - drop any keys whose value is nullish

        Returns:
            dict
        """
        outd = {}
        for name in self._field_names:
            field = self._fields.get(name)
            value = getattr(self, name)
            if hasattr(value, "serialize"):
                value = value.serialize(implicit_nulls=implicit_nulls)

            if hasattr(field, "serialize"):
                value = field.serialize(value, implicit_nulls=implicit_nulls)

            if implicit_nulls and nullish(value,
                                          implicit_nulls=implicit_nulls):
                pass
            else:
                outd[name] = value

        return outd

    def validate(self):
        """
        Validate the values in the schema.

        Returns:
            True

        Raises:
            Invalid
        """
        return all([v.validate(getattr(self, k))
                    for k, v in self._fields.items()])

    @property
    def is_empty(self):
        for name in self._field_names:
            if not nullish(getattr(self, name)):
                return False
        return True

    @property
    def _raw(self):
        out = dict(**self._raw_input)
        for name in self._field_names:
            value = getattr(self, name)
            out[name] = value
        return GenericSchema(**out)


class GenericSchema(SchemaInterface):
    """
    A generic schema to use when none is specified.

    """
    def __init__(self, **kwargs):
        """
        Args:
            as_dict (dict) - a blob from which to infer a schema
        """
        for k, v in kwargs.items():
            setattr(self, k, v)

    def __getattr__(self, name):
        return self.__dict__.get(name)

    def _serialize_value(self, value, implicit_nulls=True):
        """
        Serialize a value, recursively descending through the object to make
        sure any nested objects are also serialized.

        """
        if hasattr(value, "serialize"):
            value = value.serialize(implicit_nulls=implicit_nulls)

        if isinstance(value, (list, tuple)):
            items = []
            for i in value:
                ser = self._serialize_value(
                    i, implicit_nulls=implicit_nulls)

                if not nullish(ser, implicit_nulls=implicit_nulls):
                    items.append(ser)

            value = items

        if implicit_nulls and nullish(
                value, implicit_nulls=implicit_nulls):
            return None
        return value

    def serialize(self, implicit_nulls=True):
        """
        Represent generic schema as a dictionary.

        Kwargs:
            implicit_nulls (Bool) - drop any keys whose value is nullish

        Returns:
            dict
        """
        outd = {}
        for name, value in self.__dict__.items():
            value = self._serialize_value(value, implicit_nulls=implicit_nulls)
            if implicit_nulls and nullish(
                    value, implicit_nulls=implicit_nulls):
                pass
            else:
                outd[name] = value

        return outd

    def validate(self):
        """
        *shrug*

        """
        return True

    @property
    def is_empty(self):
        return all(nullish(v) for v in self.__dict__.values())


class Mapping(MappingInterface):
    """
    A base class for defining your mappings:

    Declare a transformation from one shape to another shape.

    Just inherit this and add some fields:

        class DogToAnimal(Mapping):
            source_schema = Dog
            target_schema = Animal

            name = Get('dogname')
            type = Const('dog')
            legs = Const(4)
            noise = Const('woof!')

    The action happens when you get an instance of your mapping:

        dog_to_animal_map = DogToAnimal()

    Then apply it to an object. You get a Schema instance:

        my_animal = dog_to_animal_map.apply(my_dog)
        type(my_animal)
        # __main__.Animal

    ... which is full of values:

        my_animal.serialize()
        # {"name": "Fido", "type": "dog", "legs": 4, "noise": "woof!}

    """
    def apply(self, blob):
        """
        Take the mapping and push a blob through it.

        Args:
            blob (dict or Schema) - the thing to transform

        Returns:
            instance of `self.target_schema` (if declared) or GenericSchema
        """
        if self.source_schema is None:
            loaded_source = blob
        elif isinstance(blob, self.source_schema):
            loaded_source = blob
        else:
            loaded_source = self.source_schema(**blob or {})

        all_attrs = self._fields.keys()
        target_dict = {}
        for attr_name in all_attrs:
            transform = self._fields[attr_name]
            result = transform(loaded_source)
            target_dict[attr_name] = result

        if self.target_schema is None:
            return GenericSchema(**target_dict)

        return self.target_schema(**target_dict)



"""
Exceptions for BFH

"""
from __future__ import absolute_import

__all__ = [
    "Invalid",
    "Missing",
]


class Invalid(TypeError):
    """
    The value in this field is not valid.

    """


class Missing(KeyError, AttributeError):
    """
    A thing that should be here... is not.

    """



"""
Utility functions for BFH

"""
from __future__ import absolute_import

import re
from datetime import timedelta, tzinfo

__all__ = [
    "NULLISH",
    "dedunder",
    "nullish",
    "utc",
]


DUNDER_MATCH = re.compile(r'_.+?__')  # python's privacy prefix "_Foo__"


def dedunder(name):
    """
    HORROR, SHAME, GASP.

    Subvert Python's very prudent double-underscore privacy mechanism.

    "_Foo__my_attr" -> "my_attr"

    Args:
        name (str): a name that may or may not have the prefix

    Returns:
        the name stripped of the privacy prefix, if it had one.
    """
    if DUNDER_MATCH.match(name):
        return DUNDER_MATCH.sub('', name)
    return name


# Types that are falsey, but not False itself.
NULLISH = (None, {}, [], tuple())


def nullish(value, implicit_nulls=True):
    """
    Kwargs:
        implicit_nulls (bool): accept empty containers as well as None
    """
    if implicit_nulls:
        if hasattr(value, 'is_empty'):
            return value.is_empty

        return value in NULLISH

    return value is None


class UTC(tzinfo):
    """
    UTC tzinfo.

    Why isn't this in the standard library?
    """
    OFFSET = timedelta(0)

    def utcoffset(self, dt):
        return self.OFFSET

    def tzname(self, dt):
        return "UTC"

    def dst(self, dt):
        return self.OFFSET

utc = UTC()



"""
Transformations that can go on Mappings.

When you `apply` a mapping to an object, the transformations
filter that object into the form of the output schema.

- You can `Get` the value from a field and pass it somewhere.
- You can use `All` to pass the whole input object somewhere.
- You can coerce a value into a type with `Bool`, `Int`, `Num`, or `Str`.
- You can `Concat` some values together into one value.
- You can use `Many` to group some values into a list.
- You can pass a constant value with `Const` no matter what the input object
- You can `Do` arbitrary functions on input.
- You can nest mappings inside others with `Submapping` and `ManySubmap`

"""
from __future__ import absolute_import

from datetime import datetime

from dateutil.parser import parse as parse_date
import six

from itertools import chain

from .common import utc
from .exceptions import Missing
from .interfaces import TransformationInterface

try:
    unicode_type = unicode
except NameError:
    unicode_type = str

__all__ = [
    "All",
    "Bool",
    "Concat",
    "Const",
    "Do",
    "Get",
    "Int",
    "Many",
    "ManySubmap",
    "Num",
    "Str",
    "Submapping",
]


class All(TransformationInterface):
    """
    Get the *whole darn source object*

    Kwargs:
        strict (Bool) - If the object is a schema, drop any extra keys
                        that don't appear in the schema.

    Returns:
        If strict == False and called on a Schema instance, a GenericSchema
        Otherwise just the object.
    """
    def __init__(self, strict=False):
        self.strict = strict

    def __call__(self, source):
        return self.function(source)

    def function(self, source):
        if not self.strict and hasattr(source, '_raw'):
            return source._raw
        return source


class Get(TransformationInterface):
    """
    Gets a value from a dict or object

        value = Get("my_thing")({"my_thing": "get this"})
        value == "get this"

    Fetch values from complex objects by passing multiple arguments.

        value = Get("a", "b")({"a": {"b": 1}})
        value == 1

    Args:
        a series of names to follow into object and subobject.

    Kwargs:
        required (bool, default False) -- error if names missing

    Returns:
        fetched value or None if `required`

    Raises:
        Missing if `required` is false

    """
    def __init__(self, *args, **kwargs):
        """
        """
        self.path = args
        self.kwargs = kwargs
        self.required = kwargs.get('required', False)

    def __call__(self, source):
        return self.function(source)

    def _get_from_dict(self, source, path):
        if not self.required:
            return source.get(path)
        return source[path]

    def _get_from_obj(self, source, path):
        if not self.required:
            return getattr(source, path, None)
        return getattr(source, path)

    def _get(self, source, path):
        try:
            if isinstance(source, dict):
                return self._get_from_dict(source, path)
            return self._get_from_obj(source, path)
        except (KeyError, AttributeError) as e:
            raise Missing(e)

    def function(self, source, path=None):
        parts = path or self.path
        first, rest = parts[0], parts[1:]
        got = self._get(source, first)
        if rest:
            return self.function(got, rest)
        return got


class Transformation(TransformationInterface):
    """
    A callable thing that transforms an input.

    Base class for transformations; they should implement `function`.

    Kwargs:
        required (bool, default False) -- error if names missing
    """
    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs
        if kwargs.get('required') is None:
            self.required = False
        else:
            self.required = kwargs.get('required')

    def __call__(self, source=None):
        call_args = []
        for arg in self.args:
            if isinstance(arg, TransformationInterface):
                call_args.append(arg(source))
            else:
                call_args.append(arg)

        return self.function(source, *call_args)


class Submapping(Transformation):
    """
    Map a complex subobject into a subschema.

    Returns:
        result of applying submapping to the input
    """
    def __init__(self, submapping_class, *args):
        self.submapping_class = submapping_class
        self.args = args

    def function(self, source, *call_args):  # source ignored
        return self.submapping_class().apply(call_args[0])


def _many_items(call_args, drop_nones=True):
    """
    Helper function for array-ish transformations.

    """
    if drop_nones:
        call_args = [i for i in call_args if i is not None]

    if len(call_args) > 1:
        return call_args

    if len(call_args) == 1 and isinstance(call_args[0], (list, tuple)):
        return call_args[0]

    elif len(call_args) == 1:
        return [call_args[0]]

    else:
        return []


class ManySubmap(Submapping):
    """
    Map an array of complex objects onto a subschema.

    Returns:
        results of applying submapping to each item in the input
    """
    def function(self, source, *call_args):  # source ignored
        return [self.submapping_class().apply(item)
                for item in _many_items(call_args)]


class Many(Transformation):
    """
    Construct an array from a series of values.

    Args:
        subtrans (Transformation): transformation to apply to all input items

    Returns:
        results of applying subtransformation to each item in the input
    """
    def __init__(self, subtrans, *args, **kwargs):
        self.subtrans = subtrans  # Transformation
        self.args = args
        self.kwargs = kwargs

    def function(self, source, *call_args):
        if isinstance(self.subtrans, Submapping):
            raise ValueError("Can't Many(Submapping). Use Manymap instead.")

        return [self.subtrans(item, **self.kwargs)()
                for item in _many_items(call_args)]


class Const(Transformation):
    """
    Return a constant value.

    """
    def function(self, source, *call_args):  # source ignored
        return call_args[0]


class CoerceType(Transformation):
    """
    A base class for transformations that use basic Python type coercion

    """

    null_types = (None,)

    @property
    def target_type(self):
        raise NotImplementedError("Specify a type")

    def function(self, source, *call_args):  # source ignored
        value = call_args[0]
        if not self.required and value in self.null_types:
            return value
        return self.target_type(value)


class Int(CoerceType):
    """
    Coerce input to an integer

    """
    target_type = int


class Num(CoerceType):
    """
    Coerce input to a floating point number

    """
    target_type = float


class Str(CoerceType):
    """
    Coerce input to a unicode string

    """
    target_type = unicode_type

    null_types = (None, "")


class Bool(CoerceType):
    """
    Coerce input to a boolean

    """
    target_type = bool


class Concat(Transformation):
    """
    Concat some strings into a single string.

    """
    def __init__(self, *args, **kwargs):
        super(Concat, self).__init__(*args, **kwargs)
        self.strict = kwargs.get('strict', False)

    def function(self, source, *call_args):  # source ignored
        if not self.strict:
            call_args = [i for i in call_args if i]
        return "".join(call_args)


class Do(Transformation):
    """
    Do an arbitrary thing to something.

        class MyMapping(Mapping):
            four = Do(lambda x: x*x, Get('two'))

        MyMapping().apply({'two': 2}).serialize()
        # {'four': 4}

    Args:
        the first positional arg to the constructor should be a callable...
        this callable is applied to the input generated by the subsequent args
    """
    def function(self, source, *call_args):  # source ignored
        return call_args[0](*call_args[1:])


class Chain(Transformation):
    """Chain an arbitrary number of iterables."""
    def function(self, source, *call_args):
        return list(chain(*call_args))


class ParseDate(Transformation):
    """
    Parse a date-ish string into a datetime object.

    """
    DEFAULT_TIMEZONE = utc

    @property
    def tz(self):
        return self.kwargs.get("tz", self.DEFAULT_TIMEZONE)

    def function(self, source, *call_args):  # source ignored
        value = call_args[0]
        if isinstance(value, int):
            date = datetime.utcfromtimestamp(value)
        elif isinstance(value, six.string_types):
            date = parse_date(value)
        else:
            raise TypeError("Could not parse %s" % value)

        if date.tzinfo is None:
            date = date.replace(tzinfo=self.tz)

        return date


class DateToIsoString(Transformation):
    """
    Turn a datetime into an ISO 8601 formatted string.

    """
    def function(self, source, *call_args):  # source ignored
        try:
            return call_args[0].isoformat()
        except (IndexError, AttributeError):
            raise ValueError("Not a datetime: %s" % (
                call_args[0] if call_args else None))



"""
Fields that can go on Schemas

Included are the common data types you'll probably need, and a generic `Field`
class so you can implement your own if you want.
"""
from __future__ import absolute_import

import re
from datetime import datetime

from .common import nullish
from .exceptions import Invalid
from .interfaces import FieldInterface, SchemaInterface

try:
    string_type = unicode
except NameError:
    string_type = str


__all__ = [
    "ArrayField",
    "BooleanField",
    "DatetimeField",
    "Field",
    "IsoDateString",
    "IntegerField",
    "NumberField",
    "ObjectField",
    "Subschema",
    "UnicodeField",
]


class Field(FieldInterface):
    """
    Base class for a field.

    You can create custom fields if you inherit this and implement
    `serialize` and `validate` as desired:

        class AngryField(Field):
            def serialize(self, value, **kwargs):
                return "I HATE YOU"

            def validate(self, value):
                return False

    """
    def __init__(self, required=True, default=None):
        """
        Initialize the field.

        Args:
            required (Bool) - is this a required field in the schema?
        """
        self.required = required
        self.default = default

    def __get__(self, instance, cls=None):
        if instance is None:
            return self

        value = instance.__dict__.get(self.field_name)
        if value is not None:
            return value
        else:
            default = self.default
            instance.__dict__[self.field_name] = default
            return default

    def __set__(self, instance, value):
        instance.__dict__[self.field_name] = value

    def serialize(self, value, **kwargs):
        """
        Transform the value into whatever shape you'd like it to appear in a
        Python dict, ready for use in the outside world.

        """
        return value

    def validate(self, value):
        """
        Is the value valid in this kind of field?

        """
        if self.required and value is None:
            raise Invalid("%s: a value is required" % self.field_name)
        return True

    @property
    def field_name(self):
        return getattr(self, "_field_name", "unnamed")

    @field_name.setter
    def field_name(self, value):
        self._field_name = value

    @property
    def default(self):
        return self._default() if callable(self._default) else self._default

    @default.setter
    def default(self, value):
        self._default = value


class Subschema(Field):
    """
    A field that defines a subschema.

    Define one schema, then you can use this field to embed it in another:

        class Inner(Schema):
            # some fields

        class MySchema(Schema):
            inner = Subschema(Inner)
            # more fields

    """
    def __init__(self, subschema_class, *args, **kwargs):
        super(Subschema, self).__init__(*args, **kwargs)
        self.subschema_class = subschema_class

    def __set__(self, instance, value):
        if isinstance(value, dict):
            instance.__dict__[self.field_name] = self.subschema_class(**value)
        else:
            instance.__dict__[self.field_name] = value

    def serialize(self, value, implicit_nulls=True):
        """
        A Subschema serializes as a dict.

        Kwargs:
            implicit_nulls (bool): drop keys with falsey values
        """
        if hasattr(value, "serialize"):
            value = value.serialize(implicit_nulls=implicit_nulls)

        if value is None:
            value = {}

        if isinstance(value, dict) and implicit_nulls:
            if all(nullish(v) for v in value.values()):
                value = {}

        return value

    def validate(self, value):
        super(Subschema, self).validate(value)
        if not self.required:
            if nullish(value):
                return True
            if getattr(value, "is_empty", False):
                return True

        if not hasattr(value, 'validate'):
            value = self.subschema_class(value)

        value.validate()
        return True


class SimpleTypeField(Field):
    """
    Base class for fields that simply validate a type.

    Expects a class attribute 'field_type'.
    """
    def _valid(self, value):
        if value is None and not self.required:
            return True
        return isinstance(value, self.field_type)

    def validate(self, value):
        super(SimpleTypeField, self).validate(value)
        if not self._valid(value):
            raise Invalid("%s: %s is not a valid %s" % (self.field_name,
                                                        value,
                                                        self.field_type))
        return True


class BooleanField(SimpleTypeField):
    """
    A field that should contain a boolean.

    """
    field_type = bool


class IntegerField(SimpleTypeField):
    """
    A field that should contain an integer.

    """
    field_type = int


class NumberField(SimpleTypeField):
    """
    A field that should contain a floating-point number.

    """
    field_type = float


class DatetimeField(SimpleTypeField):
    """
    A field that should contain a Python datetime object.

    """
    field_type = datetime


class UnicodeField(SimpleTypeField):
    """
    A field that contains strings.

    Expect unicode. If it's not already unicode, assume it's utf-8 and
    transform it to unicode.
    """
    field_type = string_type

    def __init__(self, strict=False, encoding='utf-8', **kwargs):
        """
        Kwargs:
            strict (bool): don't coerce a non-unicode string to unicode
            encoding (str): default 'utf-8'
        """
        super(UnicodeField, self).__init__(**kwargs)
        self.strict = strict
        self.encoding = encoding

    def _coerce(self, value):
        if isinstance(value, string_type):
            return value
        try:
            return value.decode(self.encoding)
        except AttributeError:
            raise Invalid("%s: %s is not a string" % (self.field_name, value))

    def validate(self, value):
        if self.strict or (value is None and not self.required):
            return super(UnicodeField, self).validate(value)
        return super(UnicodeField, self).validate(self._coerce(value))

    def serialize(self, value, **kwargs):
        try:
            return self._coerce(value)
        except Invalid:  # we are not in the business of validation here
            return value


class IsoDateString(UnicodeField):
    """
    A string field that validates that it contains an ISO 8601 date string

    """
    ISO_REGEX = re.compile(r"\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}")

    def validate(self, value):
        if not self.required and value is None:
            return True

        super(IsoDateString, self).validate(value)
        if not bool(self.ISO_REGEX.match(value)):
            raise Invalid("%s not an ISO 8601 date string")

        return True


class ObjectField(SimpleTypeField):
    """
    A field that can contain a schemaless dict or object.

    """
    field_type = (dict, SchemaInterface)

    def validate(self, value):
        super(ObjectField, self).validate(value)
        if hasattr(value, "validate"):
            value.validate()
        return True

    def serialize(self, value, implicit_nulls=True):
        """
        Kwargs:
            implicit_nulls (bool): drop keys with falsey values
        """
        if hasattr(value, "serialize"):
            value = value.serialize(implicit_nulls=implicit_nulls)

        if value is None:
            value = {}

        if isinstance(value, dict) and implicit_nulls:
            if all(nullish(v) for v in value.values()):
                value = {}

        return value


class ArrayField(SimpleTypeField):
    """
    A field that can contain an array of things of type `array_type`

        class Inner(Schema):
             wow = IntegerField()

        class MySchema(Schema):
             ints = ArrayField(int)
             inners = ArrayField(Inner)

    indicates an object that looks like so:

        {"ints": [1, 2, 3], "inners": [{"wow": 4}, {"wow": 5}]}

    """
    field_type = (list, tuple)

    def __init__(self, array_type=None, **kwargs):
        """
        Kwargs:
            array_type (type or Schema)
        """
        super(ArrayField, self).__init__(**kwargs)
        self.array_type = array_type

    def _flatten(self, value, implicit_nulls=True):
        if hasattr(value, 'serialize'):
            return value.serialize(implicit_nulls=implicit_nulls)
        return value

    @property
    def is_schema_type(self):
        return issubclass(self.array_type, SchemaInterface)

    def validate(self, items):
        # blech... it's not a validation lib. it's not a validation lib.
        super(ArrayField, self).validate(items)
        if not self.required and items in (None, [], tuple()):
            return True

        if self.is_schema_type:
            for val in items:
                if isinstance(val, self.array_type):
                    return val.validate()
                if isinstance(val, SchemaInterface):  # wrong schema
                    raise Invalid("%s is not a %s" % (val, self.array_type))

                # val is an object literal. see if it matches schema.
                in_schema = self.array_type(val)
                return in_schema.validate()

        elif self.array_type is not None and items is not None:
            for val in items:
                if not isinstance(val, self.array_type):
                    raise Invalid("%s is not a %s" % (val, self.array_type))

        return True

    def serialize(self, value, implicit_nulls=True):
        if isinstance(value, self.field_type):
            items = []
            for i in value:
                flat = self._flatten(i, implicit_nulls=implicit_nulls)
                if not nullish(flat, implicit_nulls=implicit_nulls):
                    items.append(flat)
            return items
        return value



#!/usr/bin/env python

from setuptools import setup

exec(open('bfh/version.py').read())

setup(
    name='bfh',
    version=__version__,
    description="Smacks schemas into other schemas",
    author="Evan Bender",
    install_requires=[
        "python-dateutil",
        "six",
    ],
    author_email="evan.bender@percolate.com",
    url="https://github.com/percolate/bfh",
    classifiers=[
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: Mozilla Public License 2.0 (MPL 2.0)",
        "Programming Language :: Python :: 2.7",
    ],
    packages=["bfh"],
)



"""
Interfaces for BFH

"""
from __future__ import absolute_import

from abc import ABCMeta, abstractmethod, abstractproperty
from six import add_metaclass

from .common import dedunder

__all__ = [
    "FieldInterface",
    "MappingInterface",
    "SchemaInterface",
    "TransformationInterface",
]


@add_metaclass(ABCMeta)
class FieldInterface(object):
    """
    Descriptor that defines a field on a Schema class.

    """
    @abstractmethod
    def validate(self):
        """
        Validate the field.

        """

    @abstractmethod
    def serialize(self, value, implicit_nulls=True):
        """
        Serialize the field.

        """


@add_metaclass(ABCMeta)
class TransformationInterface(object):
    """
    Descriptor that defines a field transformation on a Mapping class.

    """
    @abstractmethod
    def function(self, whole_obj, *call_args, **kwargs):
        """
        The function that is called to transform a field.

        """


class HasFieldsMeta(ABCMeta):
    """
    Metaclass for classes that may have fields.

    """
    def __new__(metaclass, classname, bases, attributes, *args, **kwargs):
        new_class = super(HasFieldsMeta, metaclass).__new__(
            metaclass, classname, bases, attributes, *args, **kwargs
        )
        setattr(new_class, '_fields', {})
        setattr(new_class, '_field_names', [])
        for name in dir(new_class):
            attribute = getattr(new_class, name)
            if not isinstance(attribute,
                              (FieldInterface, TransformationInterface)):
                continue
            name = dedunder(name)
            new_class._fields[name] = attribute
            new_class._field_names.append(name)
            attribute.field_name = name
        return new_class


@add_metaclass(HasFieldsMeta)
class SchemaInterface(object):
    """
    Interface for a Schema class.

    """
    @abstractmethod
    def validate(self):
        """
        Validate the schema instance.

        """

    @abstractmethod
    def serialize(self, implicit_nulls=True):
        """
        Serialize the schema instance.

        """

    @abstractproperty
    def is_empty(self):
        """
        Is the schema instance empty?

        """


@add_metaclass(HasFieldsMeta)
class MappingInterface(object):
    """
    Interface for a Mapping class.

    """
    @property
    def source_schema(self):
        """
        Optionally declare the source schema for this mapping.

        """

    @property
    def target_schema(self):
        """
        Optionally declare the target schema for this mapping.

        """

    @abstractmethod
    def apply(self, blob):
        """
        Apply the mapping to an object.

        """
